## Score computer for the EvAAL competition
##
## Copyright © Francesco Potortì, 2015
## Licensed to anyone under the GNU Affero General Public License 3.0

## $Revision: 2.16 $

## When this script is called from Octave it looks under BASEDIR for
## directories containing all the necessary files (see below) and shows
## them in a menu for the user to choose from.  After the choice is
## done, data are read and a report is created.  If plotting is requested,
## a graph displaying the estimated path superimposed over the map
## is shown on video and saved under the chosen dir in a pdf file.

## The normal way to call this script from the Octave prompt is:
## >> evaalscore
## or
## >> doplot=false; basedir="../competitors/data"; evaalscore
## where doplot defaults to true and basedir defaults to "."

## To plot the path only, without errors, create a directory
## "EvAAL-2015-path" containing all the necessary files but without the
## legend.png file (lack of legend.png prevents creation of title,
## legend, statistics and distribution graph), then do:
## >> dirname="EvAAL-2015-path";
## >> create_dummy_positions([basedir "/" dirname], 0); evaalscore

## The files 'buttonsPressed.log' and 'positions.log' are those created
## by the stepLogger app.  The reference system used for the coordinates
## must be either a local system with unit in meters or else WGS84. In
## the latter case, the  coordinates must be in the order long, lat. The
## file 'distance' contains one line of text: either "localdistance" or
## "wgs84distance", depending on the reference system.

## If present, the file 'map' is an image file (PNG 8 bit works, most
## other formats work too) used as the graph background, and
## 'map_corners' is a text file containing the coordinates of the four
## corners of the map in the usual image order: NO, NE, SE, SW

floorpenalty = 15;			# Floor penalty [m] used when computing distance

## Set default values
exist("basedir") || (basedir = ".");
exist("doplot") || (doplot = true);

global mapinfo spsr spsc

if (doplot)
  pkg load image
  pkg load plot
endif

## Create a random "positions.log" file useful for testing
## N is how many new pos to create between existing ones
## iF n IS 0, create perfect positions, that is, without errors
##
## Call like this:
##  create_dummy_positions ("testevaal", 5)
function  create_dummy_positions (directory, n)
  positionsfile = [directory "/positions.log"];

  t = textread([directory "/buttonsPressed.log"], "%f%*%*%*%*", 'delimiter', ":");
  [l x y f] = textread([directory "/groundTruth.txt"], "%s%f%f%f");
  if (length(t) != length(l))
    warning("%d ground truth labels, but %d button pressures\n",
	  length(l), length(t));
    return
  endif

  if (n == 0)			# copy reference points
    spd = [t-1 x y f];
  else				# interpolate and perturbate ref points
    p = interp1([t x y f], 0:1/(n+1):length(t)+2, 'extrap');
    absdiff = quantile(abs(diff(p)), 0.5)
    d = n * (rand(size(p))-1) .* absdiff;
    pd = p+d;
    pd(:,4) = round(pd(:,4));		# floor must be integer
    spd = sortrows(pd);			# sort by increasing times
  endif

  fid = fopen(positionsfile, 'w');
  fprintf(fid, "%.0f\t%.6f\t%.6f\t%d\n", spd');
  fclose(fid);
  printf("File %s written\n", positionsfile);
endfunction

## Create an IMAGE.corners file from IMAGE.png and IMAGE.points
function create_corners(image)
  imagefile = [ image ".png"];
  pointsfile = [ image ".points"];
  cornersfile = [ image ".corners"];

  ## Create coordinate conversion matrix pixel_to_ref from pointsfile
  pw = imfinfo(imagefile).Width;
  ph = imfinfo(imagefile).Height;
  corners_pixel = [0,0; pw,0; pw,-ph; 0,-ph];
  [mapX mapY pixelX pixelY] = textread(pointsfile, "%f%f%f%f%*", 'delimiter', ",", 'headerlines', 1);
  pixel_to_ref = cp2tform ([pixelX pixelY], [mapX mapY], 'affine');

  ## Create image corners cordinates and store them into cornersfile
  printf("Saving %s ...\n", cornersfile);
  mapcorners = tformfwd(pixel_to_ref, corners_pixel);
  printf("%12.6f%12.6f\n", mapcorners);
  save_precision(10, 'local');
  save("-text", cornersfile, "mapcorners");
endfunction


## Look in BASEDIR for directories containing the necessary files and
## ask the user to choose one of them, return the chosen dir as a string,
## that is the dir name relative to BASEDIR
function dir = chosendir(basedir);

  ## The files expected in a dir with one or no maps.  Optional files "map"
  ## and "map_corners" are required to add a background map
  files_single_map = { "buttonsPressed.log",
		       "positions.log",
		       "distance"};

  files = readdir(basedir);
  gooddiridx = [];
  for i = 1:length(files)
    if (isdir(files{i}))
      dirfiles = readdir(files{i});
      found_single = 0;
      for j = 1:length(dirfiles)
	found_single += any(strcmp(dirfiles{j}, files_single_map));
      endfor
      if (found_single == length(files_single_map))
	gooddiridx(end+1) = i;
      endif
    endif
  endfor
  dirs = files(gooddiridx);

  assert(length(dirs) > 0, "No good directories found under '%s, exiting\n", basedir);
  choice = 0;
  choice = menu("Choose a directory", {"no choice: EXIT" dirs{:}});
  assert(choice > 1, "No directory chosen, exiting\n");

  dir = dirs{choice-1}			# first choice is "no choice"
endfunction


## Convert vectors of X and Y coordinates to pixel coords
function [px py] = to_pixel_coords (ph, pixel_to_ref, x, y)
  p = tforminv(pixel_to_ref, [x y]);
  px = p(:,1);
  # We use negative coords for images, contrary to the usual convention
  # of y axis going downwards, and below we use flipud on the image.
  py = p(:,2) + ph;
endfunction

#
## Draw graph
## Read the data files, put the reference coordinates in RX RY and the
## estimated coordinates in EX EY, draw the estimated path with the map
## in background, if available
function [rl rt rx ry rf et ex ey ef] = errorgraph (directory, doplot)
  mapfile = [directory "/map"];
  cornersfile = [directory "/map_corners"];
  buttonsfile = [directory "/buttonsPressed.log"];
  positionsfile = [directory "/positions.log"];
  groundtruthfile = [directory "/groundTruth.txt"];

  mappresent = !isempty(stat(mapfile)) && !isempty(stat(cornersfile));
  groundtruthpresent = !isempty(stat(groundtruthfile));

  if (doplot && mappresent)
    ## Create coordinate conversion matrix pixel_to_ref
    pw = imfinfo(mapfile).Width;
    ph = imfinfo(mapfile).Height;
    corners_pixel = [0,0; pw,0; pw,-ph; 0,-ph];
    corners_ref = load('-ascii', cornersfile); # Ignore Octave headers
    pixel_to_ref = cp2tform(corners_pixel, corners_ref, 'projective');

    ## Draw the map as background
    alpha = 0.7;	    # transparency of map
    light = 0.5;	    # light level of grey behind transparent map
    map = alpha*light*255 + (1-alpha)*flipud(imread(mapfile));
    image(map);
    axis('xy', 'image', 'off', 'nolabel'); # should use 'ij' and no flipud
    hold on
  endif

  ## Read button pressed time, button label, reference coordinates (ground truth)
  [bpt bpl bpx bpy bpf] = textread(buttonsfile, "%f%s%f%f%f", 'delimiter', ":");

  assert(issorted(bpt), "times not increasing in %s", buttonsfile);

  if (groundtruthpresent)
    ## If the ground truth file is present, the bpx and bpy coordinates that
    ## we have just read are dummy ones, and we should recover them from the
    ## ground truth file; also check that the dummy values are 0
    assert(all([bpx; bpy; bpf] == 0),
	   "%s is present, but not all coordinates in %s are 0\n",
	   groundtruthfile, buttonsfile);
    [gtl gtx gty gtf] = textread(groundtruthfile, "%s%f%f%f");

    ## Check labels and floor
    gtl = strtrim(gtl); bpl = strtrim(bpl); # clean up labels
    assert(numel(gtl) == numel(unique(gtl)), "non-unique labels in %s", groundtruthfile);
    assert(numel(bpl) == numel(unique(bpl)), "non-unique labels in %s", buttonsfile);
    assert(all(round(gtf) == gtf), "non-integer floor in %s", groundtruthfile);

    ## Find indices of button pressures into ground truth by matching
    ## the button pressure labels bpl to ground truth labels gtl
    [found idx] = ismember(bpl, gtl);
    assert(all(found), "%s does not contain all labels in %s\n",
	   groundtruthfile, buttonsfile);
    assert(issorted(idx), "labels in %s not sorted as %s", buttonsfile, groundtruthfile);
    if (length(bpl) != length(gtl))
      warning("%d ground truth labels, but %d button pressures\n",
	      length(gtl), length(bpl));
    endif

    ## Set reference coordinates of button pressures
    bpx = gtx(idx); bpy = gty(idx); bpf = gtf(idx); # button pressures coordinates
    bpl = gtl(idx);				     # button pressure labels
  endif
  rl = bpl; rt = bpt;			# OUTPUT labels and time
  rx = bpx; ry = bpy; rf = bpf;		# OUTPUT reference coordinates

  ## Read estimation time, estimated coordinates (competitor's output)
  [pt px py pf] = textread(positionsfile, "%f%f%f%f");

  ## Check times and floor
  assert(issorted(pt), "times not increasing in %s", positionsfile);
  assert (all(round(pf) == pf), "non-integer floor in %s", positionsfile);

  pidx = lookup(pt, bpt, 'l');		# indices of competitor estimates
  bex = px(pidx); bey = py(pidx); bef = pf(pidx); # estimates at time of button pressed
  et = pt(pidx); ex = bex; ey = bey; ef = bef;	  # OUTPUT estimated coordinates
  assert(all(rt >= et), "internal consistency check failed");

  if (!doplot) return; endif;	# stop here after reading data files

  ## Draw the estimated path, estimated points, real points, error segments
  #########################################################################

  ## Create coordinates of error lines for plotting, NA is used to create
  ## separate segments when plotting
  errx = reshape([bpx ex NA(size(bpx))]', [], 1);
  erry = reshape([bpy ey NA(size(bpy))]', [], 1);

  if (mappresent)
    [px, py] = to_pixel_coords(ph, pixel_to_ref, px, py);
    [errx, erry] = to_pixel_coords(ph, pixel_to_ref, errx, erry);
    [bpx, bpy] = to_pixel_coords(ph, pixel_to_ref, bpx, bpy);
    [bex, bey] = to_pixel_coords(ph, pixel_to_ref, bex, bey);
  endif

  epx = bex;			 # bex is estimated path, use px for whole path
  epy = bey;			 # bey is estimated path, use py for whole path
  plot(epx, epy, "-b;estimated path;", 'linewidth', 3,
       errx, erry, "-r;errors;",
       bpx, bpy, "og;truth points;", 'markerfacecolor', 'green',
       bex, bey, "ob;estimated points;");
  hold on

  ## Add full red dots to estimated points with a wrong floor
  wfidx = (bpf != bef);			# index of bad floor estimates
  plot(bex(wfidx), bey(wfidx), "or;wrong floor;", 'markerfacecolor', 'red');

  ## Overwrite the estimated path potted above with arrows.  The head size of the
  ## arrows should be set manually using the "hsize" variable.  The plot above
  ## must nonetheless plot the estimated path so that it appears in the legend.
  depx = diff(epx); depy = diff(epy);
  dlen = norm([depx depy], 'rows');
  hsize = 2;					# good for wnlab with map
  hsize = 0.05;					# good for wnlab without map
  hsize = 0.00001;				# good for KCCI without map
  for i = 1:length(epx)-1
      if (dlen(i) > 0)
	quiver(epx(i), epy(i), depx(i), depy(i), 0,
  	       'linewidth', 3, 'color', 'blue', 'maxheadsize', hsize./dlen(i));
      endif
  endfor

  axis('equal');

  hold off

endfunction


#

spsr = 3; spsc = 4;			# subplot size;
spi_legend = 4;				# subplot index for legend
spi_distrib = 12;			# subplot index for distribution graph

## Maps info
mapinfo = struct("map", { "KCCI-L3",  "PDC-L3" # file basename
			  "KCCI-L2",  "PDC-L2"
			  "KCCI-L1",  "PDC-L1"
			},
		 "floor", {3, 3; 2, 2; 1, 1},
		 "spi", {1:2, 3	# subplot index
			 5:6, 7
			 9:10 11
			},
		 "spp", {[.00 .64 .45 .29], [.45 .64 .30 .29],
			 [.00 .32 .45 .29], [.45 .32 .30 .29],
			 [.00 .00 .45 .29], [.45 .00 .30 .29],
			},
		 "p2r", [],		# pixel to ref tform
		 "ah", []		# axis handle
		);

## Draw EvAAL 2015 graph
## Read the data files, put the reference coordinates in RX RY and the
## estimated coordinates in EX EY, draw the estimated path with the map
## in background, if available
function [rl rt rx ry rf et ex ey ef] = evaal2015graph (directory, doplot)
  global mapinfo spsr spsc

  cornersfile = [directory "/map_corners"];
  buttonsfile = [directory "/buttonsPressed.log"];
  positionsfile = [directory "/positions.log"];
  groundtruthfile = [directory "/groundTruth.txt"];

  ## Set up the maps
  alpha = 0.7;	    # transparency of map
  light = 0.9;	    # light level of grey behind transparent map

  if (doplot)
    for m = 1:numel(mapinfo)
      ## Create coordinate conversion matrix pixel_to_ref
      mapfile = [directory "/" mapinfo(m).map ".png"];
      cornersfile = [directory "/" mapinfo(m).map ".corners"];
      pw = imfinfo(mapfile).Width;
      ph = imfinfo(mapfile).Height;
      corners_pixel = [0,0; pw,0; pw,-ph; 0,-ph];
      corners_ref = load('-ascii', cornersfile); # Ignore Octave headers
      pixel_to_ref = cp2tform(corners_pixel, corners_ref, 'projective');
      mapinfo(m).p2r = pixel_to_ref; # store it in mapinfo

      ## Draw the map as background
      map = alpha*light*255 + (1-alpha)*flipud(imread(mapfile));
      ah = subplot(spsr, spsc, mapinfo(m).spi, 'position', mapinfo(m).spp);
      mapinfo(m).ah = ah;			# axis handle
      image(map);
      axis('xy', 'image', 'off', 'nolabel'); # should use 'ij' and no flipud
      hold on
    endfor
  endif

  ## Read button pressed time, button label, reference coordinates (ground truth)
  [bpt bpl bpx bpy bpf] = textread(buttonsfile, "%f%s%f%f%f", 'delimiter', ":");

  assert(issorted(bpt), "times not increasing in %s", buttonsfile);

  ## If the ground truth file is present, the bpx and bpy coordinates that
  ## we have just read are dummy ones, and we should recover them from the
  ## ground truth file; also check that the dummy values are 0
  assert(all([bpx; bpy; bpf] == 0),
	 "%s is present, but not all coordinates in %s are 0\n",
	 groundtruthfile, buttonsfile);
  [gtl gtx gty gtf] = textread(groundtruthfile, "%s%f%f%f");

  ## Check labels and floor
  gtl = strtrim(gtl); bpl = strtrim(bpl); # clean up labels
  assert(numel(gtl) == numel(unique(gtl)), "non-unique labels in %s", groundtruthfile);
  assert(numel(bpl) == numel(unique(bpl)), "non-unique labels in %s", buttonsfile);
  assert(all(round(gtf) == gtf), "non-integer floor in %s", groundtruthfile);

  ## Find indices of button pressures into ground truth by matching
  ## the button pressure labels bpl to ground truth labels gtl
  [found idx] = ismember(bpl, gtl);
  assert(all(found), "%s does not contain all labels in %s\n",
	 groundtruthfile, buttonsfile);
  assert(issorted(idx), "labels in %s not sorted as %s", buttonsfile, groundtruthfile);
  if (length(bpl) != length(gtl))
    warning("%d ground truth labels, but %d button pressures\n",
	    length(gtl), length(bpl));
  endif

  ## Set reference coordinates of button pressures
  bpx = gtx(idx); bpy = gty(idx); bpf = gtf(idx); # button pressures coordinates
  bpl = gtl(idx);				  # button pressure labels

  rl = bpl; rt = bpt;			# OUTPUT labels and time
  rx = bpx; ry = bpy; rf = bpf;		# OUTPUT reference coordinates

  ## Read estimation time, estimated coordinates (competitor's output)
  [pt px py pf] = textread(positionsfile, "%f%f%f%f");

  ## Check times and floor
  assert(issorted(pt), "times not increasing in %s", positionsfile);
  assert (all(round(pf) == pf), "non-integer floor in %s", positionsfile);

  pidx = lookup(pt, bpt, 'l');		# indices of competitor estimates
  bex = px(pidx); bey = py(pidx); bef = pf(pidx); # estimates at time of button pressed
  et = pt(pidx); ex = bex; ey = bey; ef = bef;	  # OUTPUT estimated coordinates
  assert(all(rt >= et), "internal consistency check failed");

  if (!doplot) return; endif;	# stop here after reading data files

  ## Draw the estimated path, estimated points, real points, error segments
  #########################################################################

  ## Create coordinates of error lines for plotting, NA is used to create
  ## separate segments when plotting
  errx = reshape([bpx ex NA(size(bpx))]', [], 1);
  erry = reshape([bpy ey NA(size(bpy))]', [], 1);

  ## Do the plotting, for each map do the plot for all points in the
  ## corresponding floor
  printf("Creating plots:");

  for m = 1:numel(mapinfo)
    printf([" " mapinfo(m).map]);	# show advancement
    pixel_to_ref = mapinfo(m).p2r;	# tform used for coords conversion
    ph = imfinfo(mapfile).Height;
    [ppx, ppy] = to_pixel_coords(ph, pixel_to_ref, px, py);
    [perrx, perry] = to_pixel_coords(ph, pixel_to_ref, errx, erry);
    [pbrx, pbry] = to_pixel_coords(ph, pixel_to_ref, bpx, bpy);
    [pbex, pbey] = to_pixel_coords(ph, pixel_to_ref, bex, bey);

    pepx = pbex;		 # pbex is estimated path, use ppx for whole path
    pepy = pbey;		 # pbey is estimated path, use ppy for whole path
    depx = diff(pepx); depy = diff(pepy);
    dlen = norm([depx depy], 'rows');	# used to set the arrow head sizes below

    ofidx = (rf != mapinfo(m).floor);	# index of other floor's points
    ofidx3 = repmat(ofidx', 3, 1);	# repeat each element three times
    pepx(ofidx) = pepy(ofidx) = NA;
    perrx(ofidx3) = perry(ofidx3) = NA;
    pbrx(ofidx) = pbry(ofidx) = NA;
    pbex(ofidx) = pbey(ofidx) = NA;

    msize = 2.5;			# size of markers
    lwidth = 2;				# width of thick lines
    hsize = 5;				# size of arrow heads
    plot(mapinfo(m).ah,
	 pepx, pepy, "-b;estimated path;", 'linewidth', lwidth,
	 perrx, perry, "-r;errors;",
    	 pbrx, pbry, "og;truth points;", 'markersize', msize, 'markerfacecolor', 'green',
	 pbex, pbey, "ob;estimated points;", 'markersize', msize);
    hold on

    ## Add full red dots to estimated points with a wrong floor
    wfidx = (bpf != bef);			# index of bad floor estimates
    plot(pbex(wfidx), pbey(wfidx), "or;wrong floor;",
	 'markerfacecolor', 'red', 'markersize', msize);

    ## Overwrite the estimated path potted above with arrows.  The head size of the
    ## arrows should be set manually using the "hsize" variable.  The plot above
    ## must nonetheless plot the estimated path so that it appears in the legend.
    for i = 1:length(pepx)-1
      if (dlen(i) > 0)
	quiver(pepx(i), pepy(i), depx(i), depy(i), 0, 'color', 'blue',
  	       'linewidth', lwidth, 'maxheadsize', hsize./dlen(i));
      endif
    endfor

    legend('hide');
    axis('equal');
    hold off

  endfor

  printf(" done.\n");

endfunction


## Distance in meters given longitude and latitude
## Points P1 and P2 are [long lat floor] or 3xn matrices
function d = wgs84distance(floorpenalty, p1, p2)
  lng1 = p1(:,1); lat1 = p1(:,2);
  lng2 = p2(:,1); lat2 = p2(:,2);
  dLng = lng2 - lng1;
  dLat = lat2 - lat1;

  ## Equirectangular approximation
  c = norm(pi/180*[dLng .* cosd(lat1), dLat], 'rows');
  d = 6371000 * c;			# Mean Earth radius [m]

  ## Add floor penalty
  d += floorpenalty * abs(p1(:,3) - p2(:,3));
endfunction


## Distance given coordinates in their same measurement unit
## Points P1 and P2 are [x y f] or 3xn matrices
function d = localdistance(floorpenalty, p1, p2)
  xydiff = (p2-p1)(:,1:2);		# diffs of xy coordinates
  d = norm(xydiff, 'rows');

  ## Add floor penalty
  d += floorpenalty * abs(p1(:,3) - p2(:,3));
endfunction


#
################################################################

if (! exist("dirname"))
  dirname = chosendir(basedir);
endif
directory = [basedir "/" dirname];

distancefile = [directory "/distance"];
legendfile = [directory "/legend.png"];
errorfile = [directory "/error.report"];
qgisfile = [directory "/qgis_errors.txt"];

## Get the distance type
distance = textread(distancefile, "%s"){1};

## Prevent plotting, do that at the end only
if (doplot)
  figure(1, "visible", "off");
  clf(1);
endif

## Read data, possibly draw the path with background maps,
## return the reference coordinates and the estimated coordinates
#[rl rx ry rf et ex ey ef] = errorgraph(directory, doplot);
[rl rt rx ry rf et ex ey ef] = evaal2015graph(directory, doplot);

## Compute error statistics
error = feval(distance, floorpenalty, [rx ry rf], [ex ey ef]);
thirdq = quantile(error, 3/4, 1, 7)	# third quartile of error

if (doplot)
  ## If no legend file, then do not plot title, legend, statistics and
  ## distribution graph
  legendpresent = !isempty(stat(legendfile));

  if (legendpresent)
    ## Titles and legend
    th = title(mapinfo(1,1).ah, "maps not to scale");
    set(th, 'fontname', 'arial', 'fontweight', 'light');
    th = title(mapinfo(1,2).ah, sprintf("%s (%s)", dirname, distance));
    set(th, 'interpreter', 'none');
    subplot(spsr, spsc, spi_legend, 'position', [0.75 0.75 0.20 0.20]);
    map = flipud(imread(legendfile));
    image(map);
    axis('xy', 'image', 'off', 'nolabel'); # should use 'ij' and no flipud

    ## Draw distribution graph
    subplot(spsr, spsc, spi_distrib, 'position', [0.77 0.05 0.20 0.20]);
    maxerr = 20;				# max error to plot
    errno = length(error);			# number of error
    stairs([0; sort(error)], (0:errno)/errno);
    title("distribution");
    axis([0 maxerr 0 1]);
    tics('y', 0:0.25:1);
    grid on
    hold on
    plot(thirdq*[1 1], [0 1], 'linewidth', 2, 'color', 'red');

    ## Write statistics
    s = sprintf("median: %.1f\nmean: %.1f\nrms: %.1f\n\n3^{rd} quartile: %.1f",
		median(error), mean(error), std(error, 1), thirdq);
    text(2, 2, s);
  endif

  ## Generate PDF file and then display
  pdfoutfile = [directory "/score.pdf"];
  print(pdfoutfile);
  figure(1, "visible", "on") ;

endif

## Print Qgis data
fid = fopen(qgisfile, 'w');
    fprintf(fid,"id;x;y;z;id1;x1;y1;z1\n");
for l = 1:length(rl)
    fprintf(fid,"%s;",rl{l});
    fprintf(fid,"%.6f;%6f;%d;", [rx ry rf](l,:));
    fprintf(fid,"%s",rl{l});
    if (rf(l) != ef(l)) fprintf(fid,"*") endif
    fprintf(fid,";%.6f;%6f;%d\n", [ex ey ef](l,:));
endfor
fclose(fid);

## Print error report
fid = fopen(errorfile, 'w');
fprintf(fid, "EvAAL-ETRI competition 2015, Banff (CA)\n\n");
fprintf(fid, "\t\t\t================ %s ================\n\n", dirname);
fprintf(fid, "\t\t\t  === Third quartile of error: %.1f [m] ===\n\n", thirdq);
fprintf(fid, "Floor penalty is %.1f metres\n", floorpenalty);
fprintf(fid, "Error statistics [m]: median: %.1f  mean: %.1f  RMS: %.1f\n\n",
	median(error), mean(error), std(error, 1));
fprintf(fid, "\
Label\tReference coordinates\t\tEstimated coordinates\tError[m]  Error[s]\
\n\n");
for l = 1:length(rl)
  fprintf(fid, "%5s%14.6f%10.6f%3d\t%11.6f%10.6f%3d",
	  rl{l}, [rx ry rf ex ey ef](l,:));
  if (rf(l) != ef(l))
    fprintf(fid, "%+8.1f",  error(l));
  else
    fprintf(fid, "%8.1f",  error(l));
  endif
  fprintf(fid, "%10.1f\n", (rt(l)-et(l))/1000);
endfor
fclose(fid);

## Cleanup
clear basedir dirname doplot
